package org.gnu.grub.fonttool;

class CharStorageInfo {
    private final int codePoint;
    private final int fileOffset;

    public CharStorageInfo(int codePoint, int fileOffset) {
        this.codePoint = codePoint;
        this.fileOffset = fileOffset;
    }

    public int getCodePoint() {
        return codePoint;
    }

    public int getFileOffset() {
        return fileOffset;
    }
}
