package org.gnu.grub.fonttool;

import java.awt.image.BufferedImage;
import java.util.HashMap;
import java.util.Map;

public class BDFFont {
    private Map<Character, BufferedImage> glyphs = new HashMap<Character, BufferedImage>();
    private int height = 0;

    void putGlyph(char ch, BufferedImage glyph) {
        glyphs.put(ch, glyph);
        if (glyph.getHeight() > height)
            height = glyph.getHeight();
    }

    public BufferedImage getGlyph(char c) {
        return glyphs.get(c);
    }

    public int getHeight() {
        return height;
    }
}
