/* get_disk_name.c */
/*
 *  GRUB  --  GRand Unified Bootloader
 *  Copyright (C) 2007 Free Software Foundation, Inc.
 *
 *  GRUB is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  GRUB is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GRUB.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <grub/util/misc.h>

char *
grub_util_get_disk_name (int disk, char *name  __attribute__ ((unused)))
{
  char *p;

  p = xmalloc (16);
  sprintf (p, "hd%d", disk);
  
  return p;
}
