package org.gnu.grub.fonttool;

/**
 * Section name constants for the PFF2 file format.
 */
public class PFF2Sections {
    static final String FILE = "FILE";
    static final String FONT_NAME = "NAME";
    static final String FONT_FAMILY = "FAMI";
    static final String FONT_WEIGHT = "WEIG";
    static final String FONT_SLANT = "SLAN";
    static final String FONT_POINT_SIZE = "PTSZ";
    static final String MAX_CHAR_WIDTH = "MAXW";
    static final String MAX_CHAR_HEIGHT = "MAXH";
    static final String FONT_ASCENT = "ASCE";
    static final String FONT_DESCENT = "DESC";
    static final String CHAR_INDEX = "CHIX";
    static final String REMAINDER_IS_DATA = "DATA";
}
